# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps")
"""

# Python compatibility:
from __future__ import absolute_import

__all__ = [
        ## _args (Helferlein für **kwargs):
        'extract_object_and_brain', # --> (o, brain)
        'extract_object_or_brain',  # --> (o, brain=None)
        'extract_brain_or_object',  # --> (brain, o=None)
        ## _attr:
        'make_attribute_setter',
        ## _decorator.py:
        'step',
        'importStep',
        'StepAborted',
        ## _get_object:
        'make_object_getter',
        ## _make_folder:
        'make_subfolder_creator',
        ## _query:
        'make_query_extractor',
        'iterate_query',
        'getAllLanguages',
        ## _reindex.py:
        'make_reindexer',
        'reindex_all',
        ## _rename:
        # 'make_mover',  (noch nicht implementiert)
        'make_renamer',
        'ACCEPT_ANY',
        ## _roles:
        'set_local_roles',
        'make_simple_localroles_function',
        ## _switch:
        'switch_menu_item',  # Menüeintrag (de)aktivieren
        'show_item',
        'hide_item',
        ## _tree:
        'make_subfolder_creator',
        ## _types:
        'setVersionedTypes',  # --> setVersionableContentTypes
        ## _uid:
        'make_distinct_finder',
        'make_uid_setter',
        'make_uid_collector',
        ## _watch:
        'make_watcher_function',  # --> Signatur f(brain, string)
        ## _workflow:
        'make_transition_applicator',
        'load_and_cook',  # {css,js}registry.xml
        ## created by factory:
        'safe_context_id',
        ]
# Local imports:
from visaplan.plone.tools.setup._args import (
    extract_brain_or_object,
    extract_object_and_brain,
    extract_object_or_brain,
    )
from visaplan.plone.tools.setup._attr import make_attribute_setter
from visaplan.plone.tools.setup._decorator import StepAborted, importStep, step
from visaplan.plone.tools.setup._get_object import make_object_getter
from visaplan.plone.tools.setup._gs import load_and_cook, safe_context_id
from visaplan.plone.tools.setup._make_folder import make_subfolder_creator
from visaplan.plone.tools.setup._query import (
    getAllLanguages,
    iterate_query,
    make_query_extractor,
    )
from visaplan.plone.tools.setup._reindex import make_reindexer, reindex_all
from visaplan.plone.tools.setup._rename import ACCEPT_ANY, make_renamer
from visaplan.plone.tools.setup._roles import (
    make_simple_localroles_function,
    set_local_roles,
    )
from visaplan.plone.tools.setup._switch import (
    hide_item,
    show_item,
    switch_menu_item,
    )
from visaplan.plone.tools.setup._tree import clone_tree
from visaplan.plone.tools.setup._types import setVersionedTypes
from visaplan.plone.tools.setup._uid import (
    make_distinct_finder,
    make_uid_collector,
    make_uid_setter,
    )
from visaplan.plone.tools.setup._watch import make_watcher_function
from visaplan.plone.tools.setup._workflow import \
    make_transition_applicator  # TODO: transitions_map argument!



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
Methoden zum Erzeugen von Indexfeldern;
etwas weniger "prominent" als die Funktionen aus .context
"""

# Python compatibility:
from __future__ import absolute_import

# Plone:
from plone.i18n.normalizer.base import mapUnicode

# visaplan:
from visaplan.tools.coding import safe_decode

mapping = {
138: 's',  140: 'O',  142: 'z',  154: 's',  156: 'o',  158: 'z',  159: 'Y',
192: 'A',  193: 'A',  194: 'A',  195: 'a',  196: 'A',  197: 'Aa', 198: 'E',
199: 'C',  200: 'E',  201: 'E',  202: 'E',  203: 'E',  204: 'I',  205: 'I',
206: 'I',  207: 'I',  208: 'Th', 209: 'N',  210: 'O',  211: 'O',  212: 'O',
213: 'O',  214: 'O',  215: 'x',  216: 'O',  217: 'U',  218: 'U',  219: 'U',
220: 'U',  222: 'th', 221: 'Y',  223: 's',  224: 'a',  225: 'a',  226: 'a',
227: 'a',  228: 'a',  229: 'aa', 230: 'a',  231: 'c',  232: 'e',  233: 'e',
234: 'e',  235: 'e',  236: 'i',  237: 'i',  238: 'i',  239: 'i',  240: 'th',
241: 'n',  242: 'o',  243: 'o',  244: 'o',  245: 'o',  246: 'o',  248: 'o',
249: 'u',  250: 'u',  251: 'u',  252: 'u',  253: 'y',  254: 'Th', 255: 'y'}

def getSortableTitle(context):
    """
    Normalisiere den Titel des aufrufenden Objekts;
    Ersatz für die Skin-Layer-Methode getSortableTitle.

    >>> getSortableTitle(MockContext(u'Ärmel'))
    'armel'
    """
    txt = context.Title().strip().lower()
    txt = safe_decode(txt)
    return mapUnicode(txt, mapping)


if __name__ == '__main__':
    class MockContext(object):
        def __init__(self, title):
            self.title = title
        def Title(self):
            return self.title

    # Standard library:
    import doctest
    doctest.testmod()



"""
Interactive BLOBs cleanup,
inspired by a grok.view by Mikko Ohtamaa for Plone 4.1 and Dexterity 1.1;
see:

- https://stackoverflow.com/questions/8655675/cleaning-up-poskeyerror-no-blob-file-content-from-plone-site
- https://pypi.org/project/experimental.gracefulblobmissing/

The original script depended on five.grok and deleted affected objects without
further interaction; any interesting output was written to stdout.

Here we create an HTML table first, listing the objects with broken blob
data attributes in @@check-blobs;
in a second step, @@check-blobs-delete-selected can be called to delete the
checked objects.

TODO:
- Allow to clear the affected fields only.
- move to visaplan.plone.browsers @@management
- breadcrumbs
- return messages in a list
"""

# Python compatibility:
from __future__ import absolute_import

# Standard library:
from collections import Counter
from datetime import timedelta
from time import time

# Zope:
import transaction
from Products.CMFCore.interfaces import IFolderish, IPropertiesTool
from Products.Five.browser import BrowserView
# Zope imports
from ZODB.POSException import POSKeyError
from zope.component import getMultiAdapter, queryUtility

# Local imports:
from visaplan.plone.tools._have import HAS_ARCHETYPES, HAS_BLOB, HAS_DEXTERITY
from visaplan.plone.tools.context import getMessenger

# Logging / Debugging:
from logging import getLogger

interesting_at_fields = []

if HAS_ARCHETYPES:
    # Zope:
    from Products.Archetypes.Field import FileField, ImageField
    from Products.Archetypes.interfaces import IBaseContent
    interesting_at_fields.append(FileField)
    interesting_at_fields.append(ImageField)

if HAS_BLOB:
    # Plone:
    from plone.app.blob.subtypes.file import \
        ExtensionBlobField as FileExtensionBlobField
    from plone.app.blob.subtypes.image import \
        ExtensionBlobField as ImageExtensionBlobField
    interesting_at_fields.append(FileExtensionBlobField)
    interesting_at_fields.append(ImageExtensionBlobField)

if HAS_DEXTERITY:
    # Plone:
    from plone.dexterity.content import DexterityContent
    from plone.namedfile.interfaces import INamedFile

interesting_at_fields = tuple(interesting_at_fields)
log = getLogger('fix-blobs')


class FixBlobs(BrowserView):
    """
    A management view to clean up content with damaged BLOB files

    You can call this view by

    1) Starting Plone in debug mode (console output available)

    2) Visit site.com/@@check-blobs URL

    """

    def __init__(self, *args, **kwargs):
        BrowserView.__init__(self, *args, **kwargs)
        self.count = Counter()
        self.broken = []

    # ------------------------------ [ original code by Mikko Ohtamaa ... [
    def disable_integrity_check(self):
        """  Content HTML may have references to this broken image - we cannot fix that HTML
        but link integriry check will yell if we try to delete the bad image.

        http://collective-docs.readthedocs.org/en/latest/content/deleting.html#bypassing-link-integrity-check "
        """
        ptool = queryUtility(IPropertiesTool)
        props = getattr(ptool, 'site_properties', None)
        self.old_check = props.getProperty('enable_link_integrity_checks', False)
        props.enable_link_integrity_checks = False

    def restore_integrity_check_value(self):
        """
        Restore the original value
        """
        ptool = queryUtility(IPropertiesTool)
        props = getattr(ptool, 'site_properties', None)
        props.enable_link_integrity_checks = self.old_check
    # ------------------------------ ] ... original code by Mikko Ohtamaa ]

    def check_blobs(self):
        """
        Check for broken objects and return a Python dict, featuring the keys

        'count' - a Counter {'broken': x, 'inspected': x}

        'broken' - a list of dicts ('path', 'portal_type', 'modified',
                  'created', 'title')
        """
        count = self.count
        self.log_period = 1000
        self.start = time()
        _progress = []
        if HAS_ARCHETYPES:
            log.info('Archetypes framework support found (AT)')
            _progress.append('AT: %(at_objects_broken)2d broken'
                             ' / %(at_objects_checked)6d checked')
        else:
            log.info('NO Archetypes framework support found (AT)')
        if HAS_DEXTERITY:
            log.info('Dexterity framework support found (DX)')
            _progress.append('DX: %(dx_objects_broken)2d broken'
                             ' / %(dx_objects_checked)6d checked')
        else:
            log.info('NO Dexterity framework support found (DX)')
        if _progress:
            self.progress_mask = '; '.join(_progress)
        else:
            txt = 'No supported content framework found!'
            log.error(txt)
            return {
                'count': self.count,
                'broken': self.broken,
                'error': txt,
                }
        portal = self.context
        log.info('Checking objects below %(portal)r for broken BLOBs ...',
                 locals())
        self.recursively_check(portal)
        checked_total = count['objects_checked']
        if checked_total % self.log_period:
            log.info(self.progress_mask % count)
        bogus = count['unchecked_unknown']
        if bogus:
            log.warn('%(bogus)d objects not checked (unknown framework)',
                     locals())
        broken_total = count['objects_broken']
        if broken_total:
            for i, dic in enumerate(self.broken, 1):
                self.log_broken(dic, i, broken_total)

        log.info('DONE: Checked %(checked_total)d objects below %(portal)r.',
                 locals())
        total = timedelta(seconds=float(time() - self.start))
        log.info('Total time: %(total)s', locals())
        return {
            'count': self.count,
            'broken': self.broken,
            }

    def log_broken(self, dic, i, total):
        liz = ['BROKEN [%d/%d]:' % (i, total)]
        for key in ('path', 'portal_type', 'title', 'created', 'modified'):
            val = dic.get(key)
            if val:
                liz.append('%(key)s=%(val)r' % locals())
        log.error(' '.join(liz))

    def _check_blobs_of(self, o):
        count = self.count
        cnt_key = 'objects_checked'
        try:
            if HAS_ARCHETYPES and IBaseContent.providedBy(o):
                return self._check_at_blobs_of(o)
            elif HAS_DEXTERITY and isinstance(o, DexterityContent):
                return self._check_dx_blobs_of(o)
            else:
                cnt_key = 'unchecked_unknown'
                logger.warn('Can\'t check %(o)r: unknown framework',
                            locals())
                return
        finally:
            count[cnt_key] += 1
            count['objects_total'] += 1
            if not (count['objects_total'] % self.log_period):
                log.info(self.progress_mask % count)

    def _check_at_blobs_of(self, context):
        count = self.count
        schema = context.Schema()
        fields_checked = 0
        fields_broken = 0
        url = None
        for field in schema.fields():
            name = field.getName()
            if isinstance(field, interesting_at_fields):
                fields_checked += 1
                try:
                    field.get_size(context)
                except POSKeyError:
                    field_cls = field.__class__.__name__
                    if url is None:
                        url = context.absolute_url()
                        dic = {
                            'path': context.absolute_url_path(),
                            }
                    log.error("Found damaged AT %(field_cls)s %(name)r on %(url)s",
                              locals())
                    fields_broken += 1

        count['at_objects_checked'] += 1
        count['at_fields_checked'] += fields_checked
        if fields_broken:
            count['objects_broken'] += 1
            count['at_objects_broken'] += 1
            count['at_fields_broken'] += fields_broken
            try:
                ti = context.Title()
            except Exception as e:
                log.error('Error %(e)r getting the title of %(context)r!',
                          locals())
                ti = None
            dic.update({
                'title': ti,
                'created': context.created(),
                'modified': context.modified(),  # changed
                'portal_type': context.portal_type,
                })
            self.broken.append(dic)
            return True
        else:
            return False

    def _check_dx_blobs_of(self, context):
        """
        XXX: NOT TESTED - THEORETICAL, GUIDELINING, IMPLEMENTATION
        """

        count = self.count
        fields_checked = 0
        fields_broken = 0
        url = None

        # Iterate through all Python object attributes
        # XXX: Might be smarter to use zope.schema introspection here?
        for key, value in context.__dict__.items():
            # Ignore non-contentish attributes to speed up us a bit
            if not key.startswith("_"):
                if INamedFile.providedBy(value):
                    try:
                        value.getSize()
                    except POSKeyError:
                        if url is None:
                            url = context.absolute_url()
                            dic = {
                                'path': context.absolute_url_path(),
                                }
                        log.error("Found damaged Dexterity plone.app.NamedFile"
                                  " %(key)r on %(url)s",
                                  locals())
                        fields_broken += 1

        count['dx_objects_checked'] += 1
        count['dx_fields_checked'] += fields_checked
        if fields_broken:
            count['objects_broken'] += 1
            count['dx_objects_broken'] += 1
            count['dx_fields_broken'] += fields_broken
            try:
                ti = context.Title()
            except Exception as e:
                log.error('Error %(e)r getting the title of %(context)r!',
                          locals())
                ti = None
            dic.update({
                'title': ti,
                'created': context.created(),
                'modified': context.modified(),  # changed
                'portal_type': context.portal_type,
                })
            self.broken.append(dic)
            return True
        else:
            return False

    def recursively_check(self, folder):
        """
        Traverse the Plone site and collect information about objects with
        broken blobs
        """
        for id, child in folder.contentItems():

            self._check_blobs_of(child)

            if IFolderish.providedBy(child):
                self.recursively_check(child)

    def delete_selected(self):
        """
        Expect a list of paths in the 'delpath' variable;
        The given objects are checked again and, if indeed faulty,
        are deleted.
        """
        count = self.count = Counter()
        request = self.request
        delinquents = request.get('delpath', [])
        portal = self.context
        message = getMessenger(portal)
        notfound = []
        if not delinquents:
            message('Nothing to do', 'error')
            return count

        self.disable_integrity_check()
        done = False
        try:
            for delpa in delinquents:
                try:
                    spec = delpa
                    if spec.startswith('/'):
                        spec = spec.lstrip('/')
                    if not spec:
                        log.error('Invalid specification: %(delpa)r', locals())
                        count['invalid'] += 1
                        continue
                    o = portal.restrictedTraverse(spec)
                    if self._check_blobs_of(o):
                        # Yup, we have an error here:
                        o_pa = o.aq_parent
                        o_id = o.getId()
                        o_pa.manage_delObjects([o_id])
                        log.info('%(o_pa)r: deleted %(o)r', locals())
                        count['deleted'] += 1
                    else:
                        count['not_deleted'] += 1
                except KeyError:
                    notfound.append(delpa)
                    count['not_found'] += 1
            transaction.commit()
            done = True
        finally:
            self.restore_integrity_check_value()
            if done:
                log.info('Completed normally.')
            if count['deleted']:
                message('Checked and deleted ${deleted} object(s)',
                        mapping=count)
            if count['not_deleted']:
                message('No errors found for ${not_deleted} object(s)',
                        'warning',
                        mapping=count)
            if count['not_found']:
                message('${not_found} object(s) not found (already deleted?)',
                        'warning',
                        mapping=count)
            return count





# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et tw=79
"""\
functions: very simple functions for Plone
"""

# Python compatibility:
from __future__ import absolute_import

from six import text_type as six_text_type
from six.moves import map

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"
VERSION = (1,
           1,
           )
__version__ = '.'.join(map(str, VERSION))
__all__ = ['is_uid_shaped',
           'id_of',
           'short_hostname',  # expects an .absolute_url() style URL
           # ---------------- [ aus Products.unitracc.tools.misc ... [
           'looksLikeAUID',  # weniger empfohlen ...
           # ---------------- ] ... aus Products.unitracc.tools.misc ]
           ]

# Standard library:
from os.path import split, splitext

UIDCHARS_UNICODE = frozenset(u'0123456789abcdef')
def is_uid_shaped(s, onerror='raise'):
    """
    Sieht der übergebene String aus wie eine UID?

    >>> is_uid_shaped('abc123')
    False
    >>> is_uid_shaped('0123456789abcdef0123456789abcdef')
    True
    >>> is_uid_shaped('0123456789ABCDEF0123456789ABCDEF')
    False

    Verwendet z. B. zum Erzeugen des 'uid'-Werts von
    visaplan.plone.tools.forms.uid_or_number.

    ACHTUNG:
    Diese Funktion reagiert qua Vorgabe ungnädig auf unzulässige Werte:
    >>> is_uid_shaped(None)
    Traceback (most recent call last):
      ...
    ValueError: String expected: None
    >>> is_uid_shaped(42, onerror="raise")
    Traceback (most recent call last):
      ...
    ValueError: String expected: 42

    Durch Angabe eines anderen Werts als "raise" für die `onerror`-Option
    läßt sich dies beheben; *welcher* Wert dabei verwendet wird, wird aktuell
    noch ignoriert:
    >>> is_uid_shaped(None, '')
    False
    >>> is_uid_shaped(None, False)
    False

    Die Angabe "falschiger" Werte wird in diesem Fall dringend empfohlen;
    diese könnten zukünftig auch als Rückgabewert für diesen Fall verwendet
    werden.
    Wie "wahre" Strings außer 'raise' und sonstige "wahre" Werte
    zukünftig verstanden werden, ist undefiniert!

    Finally, a few basic assumptions of the function:
    >>> isinstance(b'123', bytes)
    True
    >>> isinstance(b'123', six_text_type)
    False
    >>> isinstance(u'123', six_text_type)
    True
    >>> isinstance(u'123', bytes)
    False

    The following test, however:
    >>> isinstance('123', str)
    True

    doesn't tell about the existence of the .decode method consistently in
    Python 2 and Python 3.
    """
    if isinstance(s, six_text_type):
        pass
    elif isinstance(s, bytes):
        try:
            s = s.decode('ascii')
        except UnicodeDecodeError:
            return False
    elif onerror == 'raise':
        raise ValueError('String expected: %(s)r'
                         % locals())
    else:
        return False
    return len(s) == 32 and UIDCHARS_UNICODE.issuperset(s)


# --------------------------- [ aus Products.unitracc.tools.misc ... [
UIDCHARS_BYTES = set('abcdef0123456789')
def looksLikeAUID(uid):
    """
    Sieht der übergebene String wie eine UID aus?

    >>> looksLikeAUID('abcdabcdabcdabcdabcdabcdabcdabcd')
    True
    >>> looksLikeAUID('abcdabcdabcdabcdabcdabcdabcdabc')
    False
    >>> looksLikeAUID(u'abcdabcdabcdabcdabcdabcdabcdabcd')
    True
    >>> looksLikeAUID(u'ä')
    False
    >>> looksLikeAUID('ABCDABCDABCDABCDABCDABCDABCDABCD')
    False
    >>> looksLikeAUID(42)
    False

    Die vorliegende Funktion "fällt auf Listen herein":
    >>> sillyuid = list('abcd'*8)
    >>> sillyuid[:8]
    ['a', 'b', 'c', 'd', 'a', 'b', 'c', 'd']
    >>> looksLikeAUID(sillyuid)
    True

    Sie wird hier daher vorwiegend aus historischen Gründen erhalten;
    die Umstellung auf is_uid_shaped (siehe oben) wird empfohlen.

    Es gibt allerdings einen Unterschied in der Behandlung von Fehlern,
    der dabei zu beachten ist:

    >>> looksLikeAUID(None)
    False
    """
    try:
        if len(uid) != 32:
            return False
    except TypeError:
        return False
    else:
        return set(uid).issubset(UIDCHARS_BYTES)
# --------------------------- ] ... aus Products.unitracc.tools.misc ]


def id_of(name):
    """
    Zur Verwendung in setup-Skripten etc.:
    erlaubt die Angabe einer (gf-verlinkbaren) Datei, von der für den
    auszuführenden Code nur die ID benötigt wird; z.B.:

    >>> id_of('skins/betonquali/bq_meinbetonquali.pt')
    'bq_meinbetonquali'
    >>> id_of('bq_nachrichten.pt')
    'bq_nachrichten'
    >>> id_of('fancy.xml.pt')
    'fancy.xml'
    """
    return splitext(split(name)[1])[0]


def short_hostname(url):
    """
    For use in title suffixes: we like to have crisp titles for search engines!

    >>> short_hostname('https://www.unitracc.com/some/quite/lengthy/path')
    u'unitracc.com'
    """
    liz = url.split(u'/', 3)
    hostname = liz[2]
    if hostname.startswith(u'www.'):
        return hostname[4:]
    return hostname


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: tw=79 cc=+1 sw=4 sts=4 et si
"""
visaplan.plone.tools: attools: Archetypes-related tools (_at3)
"""

def instance_and_label(instance, brain, verbose):
    if instance is None:
        instance = brain.getObject()
        if verbose:
            iinfo = brain.getURL()
    elif verbose:
        iinfo = repr(instance)
    else:
        iinfo = None
    return instance, iinfo


def getter_name(a):
    """
    >>> getter_name('userId')
    'getUserId'
    """
    return 'get%s%s' % (a[0].upper(), a[1:])


def getter_tuple(a):
    """
    Gib ein 2-Tupel aus Feldname und Getter-Name zurück
    """
    return (a, getter_name(a))


def setter_name(a):
    """
    >>> setter_name('userId')
    'setUserId'
    """
    return 'set%s%s' % (a[0].upper(), a[1:])


def setter_tuple(a):
    """
    Gib ein 2-Tupel aus Feldname und Setter-Name zurück
    """
    return (a, setter_name(a))



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _tree
"""

# Python compatibility:
from __future__ import absolute_import

# Standard library:
import random
from collections import Counter, defaultdict
from copy import deepcopy
from functools import wraps
from posixpath import normpath
from string import capitalize
from time import time

# Zope:
import transaction
from Products.CMFCore.utils import getToolByName

# Plone:
from plone.uuid.interfaces import IUUID

# visaplan:
from visaplan.tools.batches import batch_tuples
from visaplan.tools.classes import StackOfDicts
from visaplan.tools.minifuncs import gimme_False
from visaplan.tools.sequences import unique_union

# Local imports:
from .._idxnames import getExcludeFromNav
from visaplan.plone.tools._have import HAS_SUBPORTALS, HAS_VPSEARCH
from visaplan.plone.tools.setup._args import (
    _extract_move_args,
    apply_move_order_options,
    extract_layout_switch,
    normalize_menu_switch,
    setdefault_move_args,
    setdefault_source_language,
    )
from visaplan.plone.tools.setup._exc import \
    AlreadyTranslated  # from LinguaPlone, or a dummy
from visaplan.plone.tools.setup._exc import \
    CantAddTranslationReference  # ... enhanced information
from visaplan.plone.tools.setup._get_object import make_object_getter
from visaplan.plone.tools.setup._make_folder import make_subfolder_creator
from visaplan.plone.tools.setup._reindex import make_reindexer

if HAS_SUBPORTALS:
    # Local imports:
    from visaplan.plone.tools.setup._o_tools import (
        handle_subportal,
        might_set_subportal,
        )

if HAS_VPSEARCH:
    # Local imports:
    from visaplan.plone.tools.setup._o_tools import handle_united_search

# Logging / Debugging:
import logging
from pdb import set_trace
from visaplan.tools.debug import pp

__all__ = [
        'clone_tree',  # sprachverknüpfter Verzeichnisbaum
        # internal:
        # '_clone_tree_inner'
        # '_move_objects'
        ]


def clone_tree(context, dic, **kwargs):  # --- [ clone_tree ... [
    """
    Erzeuge einen sprachverknüpften Klon

    Beispiel für dic: siehe visaplan.plone.structures.setupdata
    [JOURNAL_TREE, ELEARNING_TREE, KNOWHOW_TREE]

    Erforderliche Argumente:
    - context -- üblicherweise das Portalobjekt, oder ein anderes Objekt des
                 Objektbaums (um das Portalobjekt und die Tools zu beschaffen)
    - dic -- ein Python-Dict. mit Schlüssel für die Quellsprache und eine oder
             mehr Zielsprachen sowie (optional) "children"
    - src_lang - Sprache der vorhandenen zu klonenden Struktur;
                 ein Sprachkürzel und natürlich ein Schlüssel von <dic>

    Benannte Argumente:
    - switch_menu - sollen die neuen Ordner im Menü auftauchen?
                    (Vorgabe: True)
    - move_children - sollen die Kindobjekte in das jeweilige neue
                      sprachverknüpfte Verzeichnis verschoben werden?
    - move_limit - für Test/Entwicklung: max. Anzahl der insgesamt zu
                   verschiebenden Objekte, Vorgabe: None
    - move_limit_each - dto.; max. Anzahl pro behandeltem Containerverzeichnis.
    - rectify_moved - Subportal-Angaben korrigieren für schon zuvor verschobene
                      Kindobjekte
    - create_level - 1 (Vorgabe): erzeuge fehlende Sprachvarianten,
                     2: ... auch wenn der Referenzordner fehlt;
                     3: erzeuge auch etwa fehlende Referenzordner.
    - skip_unknown_languages - sollen Zielsprachen, die in der Plone-Instanz
                     nicht aktiviert sind, übergangen werden?
                     (Vorgabe: True)

    Hinweise:
    - Es ist möglich und sinnvoll, die Funktion mit denselben Eingabedaten erst
      ohne, dann mit move_children=True aufzurufen.
    - Ein globales move_limit=0 führt dazu, daß der Verschiebungs-Code ausgeführt,
      aber wegen Überschreitung des Limits nichts verschoben wird
    - Ein move_limit_each=0 bedeutet hingegen, daß keine Obergrenze je
      Verschiebungsvorgang existiert (wohl aber ggf. ein globales move_limit,
      das natürlich beachtet würde)
    - Jede Angabe von move_limit oder move_limit_each führt dazu, daß der
      Vorgabewert für move_children True ist (siehe _args._extract_move_args)

    Diese Funktion arbeitete rekursiv:

    1. zunächst kümmert sich die veröffentlichte Funktion clone_tree um die
       Auswertung der Optionen;
    2. dann wird rekursiv _clone_tree_inner aufgerufen. Jeder Aufruf verabeitet
       ein Teil-Dictionary mit folgenden Schlüsseln:

       'de', 'en' ... *(Sprachcodes)* - die Objektangaben je Sprache
       `None` - ein optionales Dict. für Optionen
       'children' - Eine Liste mit weiteren Verästelungen

       Die Optionen werden in einer Datenstruktur StackOfDicts verwaltet;
       so ist es möglich, an die Hauptfunktion `clone_tree` in Teilbäumen zu
       übersteuern.

    """
    # We'll consume this dictionary!
    # (We don't expect it to contain much data, anyway.)
    dic = deepcopy(dic)

    pop = kwargs.get
    setdefault = kwargs.setdefault
    # locally used options:
    if 'logger' in kwargs:
        logger = pop('logger')
    else:
        logger = kwargs.setdefault(
                    'logger',
                    logging.getLogger('clone_tree:%08x'
                                      % random.randint(0, 16**8)))
    verbose = setdefault('verbose', 1)
    create_level = setdefault('create_level', 1)
    debug = setdefault('debug', False)

    # Options for get_object; set, if given:
    setdefault('set_title', True)
    setdefault('set_language', True)
    set_canonical = setdefault('set_canonical', None)
    src_lang = setdefault_source_language(kwargs)
    if not src_lang:
        raise ValueError('Currently a source_language is needed!')
    if src_lang not in dic:
        raise ValueError('input dict. lacks data'
                ' for the source language %(src_lang)r'
                % locals())

    setdefault('get_uid', True)
    setdefault('get_layout', True)

    if 'portal' in kwargs:
        portal = kwargs['portal']
    else:
        portal = getToolByName(context, 'portal_url').getPortalObject()
        kwargs['portal'] = portal
    if 'allowed_languages' in kwargs:
        allowed_languages = kwargs['allowed_languages']
    else:
        portal_languages = getToolByName(context, 'portal_languages')
        allowed_languages = portal_languages.getSupportedLanguages()
        kwargs['allowed_languages'] = allowed_languages
    logger.info('allowed languages are %(allowed_languages)r', locals())
    if src_lang not in allowed_languages:
        logger.warn('Source language %(src_lang)r is not listed as an allowed one!', locals())
    setdefault('skip_unknown_languages', not allowed_languages)

    normalize_menu_switch(kwargs)

    opt = StackOfDicts(kwargs, checked=0)
    info_collector = {
            'finally_reindex': [],
            'counter': Counter(),
            }
    try:
        return _clone_tree_inner(context, dic, opt, info_collector, {}, 0)
    finally:
        transaction.commit()
        finally_reindex = info_collector['finally_reindex']
        counter = info_collector['counter']
        pp(counter=counter)
        if finally_reindex:
            total = len(finally_reindex)
            i = 0
            for o in finally_reindex:
                i += 1
                logger.info('Reindexing %(i)d/%(total)d: %(o)r ...', locals())
                o.reindexObject()
                transaction.commit()
        errors = counter.pop('errors', 0)
        for key, val in counter.items():
            logger.info('  %(key)s: %(val)d', locals())
        if errors:
            logger.error('%(errors)d errors total', counter)
    # -------------------------------------------------- ] ... clone_tree ]


def _clone_tree_inner(context,  # --------------- [ _clone_tree_inner ... [
                      dic,  # the consumed data definition
                      opt,  # a StackOfDicts
                      info_collector,  # globally cumulated information
                      parents_o,  # `siblings_o` dict from above
                      recursion_level):
    """
    The working horse for the clone_tree function (recursive)
    """
    logger = opt['logger']

    # -------------------------- [ options dictionary processing ... [
    local_options = dic.pop(None, {})
    if local_options.get('debug'):
        logger.info('_clone_tree_inner: local debug option found; will fire again below', locals())
        set_trace()
    elif dic.get('debug'):
        logger.info('_clone_tree_inner: inherited debug option found; may fire again below', locals())
        set_trace()

    opt.push(local_options)  # local options overrides from dict

    # currently we still demand a source_language
    src_lang = opt['source_language']
    src_dict = dic.pop(src_lang)
    children = dic.pop('children', [])

    siblings_o = {}   # used during recursion
    siblings_opt = {  # options
        src_lang: src_dict.pop(None, {}),
        }
    # -------------------------- ] ... options dictionary processing ]

    counter = info_collector['counter']
    finally_reindex = info_collector['finally_reindex']

    # ----------------------------- [ _clone_tree_inner: options ... [
    create_level = opt['create_level']
    logger.info('Create_level is %(create_level)r', locals())
    create_all = create_level >= 3
    if create_all:
        logger.info('Will create all specified folders,'
                ' including missing reference folders')
    create_orphans = create_level >= 2
    if create_orphans:
        logger.info('Will create all language variants'
                ' even for missing reference folders')
    create_normal = create_level >= 1
    if create_normal:
        logger.info('Will create all specified language variants')
    else:
        logger.info("Won't create any folders")
    portal = opt['portal']
    catalog = getToolByName(portal, 'portal_catalog')
    # ----------------------------- ] ... _clone_tree_inner: options ]
    idxs = [
        getExcludeFromNav,  # changed to 'exclude_from_nav'
        'Language',
        # UNITRACC-spezifisch; visaplan.plone.subportals:
        'get_sub_portal',
        ]

    new_folder = make_subfolder_creator(logger=opt['logger'],
                                        parent=portal,
                                        idxs=idxs)
    get_object = make_object_getter(portal,
                                    logger=opt['logger'],
                                    set_title=opt['set_title'],
                                    set_language=opt['set_language'],
                                    set_canonical=opt['set_canonical'],
                                    get_uid=True,
                                    get_layout=opt['get_layout'],
                                    set_subportal=opt.get('set_subportal'),
                                    subportal=opt.get('subportal'),
                                    return_tuple=True,
                                    verbose=2)

    errors = 0

    if opt.get('debug'):
        logger.info('_clone_tree_inner: initialization done; debug option found in %(opt)r', locals())
        # pp(opt=dict(opt))
        set_trace()

    move_children, move_limit, move_limit_each = _extract_move_args(opt, do_pop=0)
    if move_children:
        reindex = False  # Container erst am Schluß reindizieren
    else:
        reindex = None

    try:
        pp(src_dict=src_dict, recursion_level=recursion_level, parents_o=parents_o)
        if src_lang in parents_o:
            src_dict['parent'] = parents_o[src_lang]
        pp(reindex=reindex, src_dict=src_dict)
        src_o, info = get_object(reindex=reindex,
                                 **src_dict)
        if src_o is None:
            specs = info['specs']
            raise ValueError('source path %(specs)s (%(src_lang)r) NOT FOUND!'
                             % locals())
        else:
            siblings_o[src_lang] = src_o
        src_dict.update(info['updates'])
        # if processing the site root, we'll not have a UID:
        # assert src_dict['uid']

        # q&d; we might want a smarter comparison method which tells
        # whether 2 dicts specify the same object (e.g. by uid, if present):
        src_path = src_dict.get('path', None)
        logger.info('source path is %(src_path)r (%(src_lang)r, %(src_o)r)',
                    locals())
        # transaction.begin()  # ist das schlau bzw. nötig?!
        if info['changes'] and not info['reindexed']:
            finally_reindex.append(src_o)

        # ... alle verbleibenden Schlüssel sind nun Sprachkürzel!

        # Wir wissen noch nicht, ob alle Kinder im selben Elternobjekt
        # verbleiben:
        same_parent = None
        canonical = None

        # --------------- [ Schleife über die *anderen* Sprachen ... [
        for la in dic.keys():
            if la not in opt['allowed_languages']:
                if opt['skip_unknown_languages']:
                    logger.error('SKIP: Clone language %(la)r is '
                                 'not listed as an allowed one!', locals())
                    continue
                else:
                    logger.warn('Clone language %(la)r is '
                                'not listed as an allowed one!', locals())
            dest_dict = dic[la]
            siblings_opt[la] = dest_dict.pop(None, {})
            if la in parents_o:
                dest_dict['parent'] = parents_o[la]
            if ('layout' in src_dict
                    and src_dict['layout']
                    and 'layout' not in dest_dict
                    and dest_dict.get('set_layout') is not False):
                logger.debug('Taking layout=%(layout)r from source folder', src_dict)
                dest_dict.update({
                    'layout': src_dict['layout'],
                    'set_layout': True,
                    })

            # -- [ gemeinsamer Container für alle Sprachen? ... [
            dest_o, info = get_object(**dest_dict)
            if dest_o is None:
                # das Elternobjekt für diese Sprache gibt es noch nicht,
                # also ist es ein anderes!
                if same_parent:
                    logger.error('We have same_parent=%(same_parent)r, '
                            'but the spec. for %(la)r (%(dest_dict)r) '
                            "doesn't exist yet!",
                            locals())
                    errors += 1
                    continue
                elif same_parent is None:
                    same_parent = False
                if create_normal:
                    dest_o = new_folder(**dest_dict)
                elif debug:
                    pp(dest_dict=dest_dict)
                    retry = 1; set_trace()
                    if retry:
                        dest_o = new_folder(**dest_dict)
            else:
                dest_dict.update(info['updates'])
                if same_parent is None:
                    same_parent = dest_dict['uid'] == src_dict['uid']
                elif same_parent:
                    if dest_dict['uid'] != src_dict['uid']:
                        dest_uid = dest_dict['uid']
                        src_uid = src_dict['uid']
                        logger.error('We have same_parent=%(same_parent)r, but '
                                'the %(la)r container object %(dest_uid)r '
                                'mismatches '
                                'the %(src_lang)r container object %(src_uid)r!',
                                locals())
                        errors += 1
                        continue
                else:
                    if dest_dict['uid'] == src_dict['uid']:
                        dest_uid = dest_dict['uid']
                        logger.error('We have same_parent=%(same_parent)r, but '
                                'the both %(src_lang)r and %(la)r '
                                'use the same container object'
                                ' %(dest_uid)r!',
                                locals())
                        errors += 1
                        continue
            siblings_o[la] = dest_o
            assert same_parent is not None
            # -- ] ... gemeinsamer Container für alle Sprachen? ]
            if same_parent:
                given_src_lang = src_dict.get('language')
                if given_src_lang:
                    logger.error('We have same_parent=%(same_parent)r, but '
                            'the source parent has language=%(given_src_lang)r '
                            'specified! (%(src_dict)s)', locals())
                    raise ValueError('Language specs')
            elif canonical is None:
                # Erster Schleifendurchlauf (erste "Zweitsprache"):
                canonical = src_o.getCanonical()  # Sprachverknüpfung
                logger.info('%(src_o)r.getCanonical() --> %(canonical)r', locals())
                src_dict.update({
                    'canonical': canonical,
                    'language': src_lang,
                    })
                info = get_object(reindex=reindex, **src_dict)[1]
                if info['changes']:
                    logger.info('Source container was changed (%(src_o)r)', locals())
            if same_parent:
                dest_dict.update({
                    'language': '',
                    })
            else:
                if opt['set_language']:
                    dest_dict.update({
                        'language': la,
                        })
                if opt['set_canonical']:
                    dest_dict.update({
                        'canonical': canonical,
                        })
            o, info = get_object(**dest_dict)
            if info['changes']:
                logger.info('Dest. container for %(la)r was changed (%(dest_o)r)', locals())
        # --------------- ] ... Schleife über die *anderen* Sprachen ]

        # -------------------- [ unitracc; visaplan.plone.search ... [
        if HAS_VPSEARCH and opt.get('united_search'):
            handle_united_search(siblings_o, opt, src_lang)
        # -------------------- ] ... unitracc; visaplan.plone.search ]

        for la, dest_dict in dic.items():
            opt.push(siblings_opt[la])

            try:
                # --- [ unitracc-spezifisch: Subportal korrigieren ... [
                rectify_moved = opt.get('rectify_moved', False)
                if rectify_moved:
                    doit = True
                    subportal = opt.get('subportal')
                    if not subportal:
                        logger.warn('rectify_moved=%(rectify_moved)r, but no subportal!', dict(opt))
                        doit = False

                    child_set_subportal = opt.get('child_set_subportal', False)
                    if not child_set_subportal:
                        logger.warn('rectify_moved=%(rectify_moved)r,'
                                    ' but not child_set_subportal!', dict(opt))
                        doit = False

                    if doit:
                        force_reindex = opt.get('force_reindex')
                        reindex = make_reindexer(logger=logger, catalog=catalog,
                                                 idxs=idxs,
                                                 # getSubPortals will change:
                                                 update_metadata=True)
                        tup = o.getPhysicalPath()
                        root_path = '/'.join(tup)
                        query = dict(path=root_path, Language=la)
                        apply_move_order_options(query, opt, do_pop=0)
                        batch_kw = {'batch_size': 100,
                                    'thingies': 'checking subportal for objects',
                                    }
                        # ---------- [ TODO: vereinheitlichen ... [
                        # (im einen Fall kommt der Wert aus der Funktion
                        # (child_set_subportal), im anderen muß er selbst
                        # gebaut werden. Besser eine von v.p.subportals
                        # erzeugte Funktion verwenden, die sich um alles
                        # kümmert und einen Wahrheitswert (geändert?)
                        # zurückgibt.)
                        if callable(child_set_subportal):
                          for batch, txt in batch_tuples(catalog(**query),
                                                         **batch_kw):
                            changes_here = 0
                            logger.info(txt + ' ...')
                            for brain in batch:
                                if brain.getPath() == root_path:  # not a child!
                                    continue
                                ch, val = child_set_subportal(brain=brain)
                                if ch or force_reindex:
                                    child_o = brain.getObject()
                                if ch:
                                    child_o.setSubPortals(val)
                                    counter['subportal_fixed'] += 1
                                    changes_here += 1
                                else:
                                    counter['subportal_checked'] += 1
                                if ch or force_reindex:
                                    if reindex(o=child_o):
                                        counter['children_reindexed'] += 1
                            if changes_here:
                                logger.info('%(changes_here)d changes; '
                                            'committing transaction ...',
                                            locals())
                                transaction.commit()
                        else:
                          for batch, txt in batch_tuples(catalog(**query),
                                                         **batch_kw):
                            changes_here = 0
                            logger.info(txt + ' ...')
                            for brain in batch:
                                if brain.getPath() == root_path:  # not a child!
                                    continue
                                val = brain.getSubPortals
                                ch = subportal not in val
                                if ch or force_reindex:
                                    child_o = brain.getObject()
                                if ch:
                                    child_o.setSubPortals(val + (subportal,))
                                    counter['subportal_fixed'] += 1
                                    changes_here += 1
                                else:
                                    counter['subportal_checked'] += 1
                                if ch or force_reindex:
                                    if reindex(o=child_o):
                                        counter['children_reindexed'] += 1
                            if changes_here:
                                logger.info('%(changes_here)d changes; '
                                            'committing transaction ...',
                                            locals())
                                transaction.commit()
                        # ---------- ] ... TODO: vereinheitlichen ]
                # --- ] ... unitracc-spezifisch: Subportal korrigieren ]

                if move_children:
                    src_child_o = siblings_o.get(src_lang)
                    if not src_child_o:
                        logger.error('Moving impossible because of missing '
                                     'source folder!')
                        errors += 1
                        move_children = False

                if move_children:
                    dest_child_o = siblings_o.get(la)
                    if dest_child_o is None:
                        logger.error('Moving impossible '
                                     'because of missing destination folder! '
                                     '(%(specs)s)',
                                     info)
                        errors += 1
                        continue

                    child_kwargs = {
                        'move_limit': move_limit,
                        'move_limit_each': move_limit_each,
                        'depth': 1,
                        }
                    subportal = opt.get('subportal')
                    set_subportal = opt.get('child_set_subportal')
                    if subportal and set_subportal:
                        child_kwargs.update({
                            'subportal': subportal,
                            'set_subportal': set_subportal,
                            })
                    move_types = opt.get('move_types') or []
                    for portal_type in move_types:
                        _move_objects(src_child_o, dest_child_o,
                                      portal_type, la,
                                      logger, counter,
                                      **child_kwargs)

                    take_types = opt.get('take_types') or []
                    for portal_type in take_types:
                        _move_objects(src_o, dest_child_o,
                                      portal_type, la,
                                      logger, counter,
                                      **child_kwargs)
                    transaction.commit()
            finally:
                opt.pop()  # (a StackOfDicts)

        # Wir haben jetzt src_o, dest_o und same_parent;
        # nun können wir in dest_o die neuen sprachverknüpften Kindelemente
        # suchen bzw. neu erstellen

        # nun die Unterordner:
        for child_dict in children:
            pp(('siblings_o:', siblings_o), ('child_dict:', child_dict))
            _clone_tree_inner(context, child_dict, opt, info_collector,
                              siblings_o,
                              recursion_level+1)

    finally:
        if errors:
            logger.error('%(errors)d errors here', locals())
            counter['errors'] += errors

    # ------------------------------------------- ] ... _clone_tree_inner ]


def _move_objects(from_o, to_o,  # ------------------ [ _move_objects ... [
                  portal_type, lang, logger, cnt,
                  **kwargs):
    """
    Helper for clone_tree(move_children)
    """
    kwargs = dict(kwargs)
    pop = kwargs.pop
    depth = pop('depth', 1)
    assert depth == 1, (
        'depth=%(depth)r: other depths than 1 are not yet implemented!'
        ) % locals()
    query = {
        'portal_type': portal_type,
        'Language': lang,
        }
    query_path = '/'.join(from_o.getPhysicalPath())
    if depth is None:
        query['path'] = query_path
        raise NotImplemented
    else:
        query['path'] = {
                'query': query_path,
                'depth': depth,
                }
    apply_move_order_options(query, kwargs)

    catalog = getToolByName(from_o, 'portal_catalog')
    brains = catalog(query)
    count_here = len(brains)
    if not count_here:
        logger.info('No %(portal_type)r objects of language %(lang)r in %(query_path)r', locals())
        return 0
    logger.info('Found %(count_here)d %(portal_type)r objects of language %(lang)r in %(query_path)r', locals())
    move_limit = pop('move_limit', None)
    move_limit_each = pop('move_limit_each', None)
    if move_limit is not None and cnt['moved_total'] >= move_limit:
        logger.info('Total move limit exceeded (%(move_limit)r)', locals())
        return 0

    try:
        if HAS_SUBPORTALS:
            consider_sp = might_set_subportal(kwargs)  # unitracc-specific
            one_by_one = portal_type == 'Folder' or consider_sp
        else:
            consider_sp = False
            one_by_one = portal_type == 'Folder'
        if one_by_one:
            i = 0
            for brain in brains:
                i += 1
                if move_limit_each and i > move_limit_each:
                    logger.info('Local move limit exceeded (%(move_limit_each)r)', locals())
                    return move_limit_each

                to_move = brain.getId
                if consider_sp:  # unitracc-specific
                    o = brain.getObject()
                    handle_subportal(o, kwargs, created=False, do_pop=False)
                    # we don't really care whether it was changed; it will be
                    # reindexed after moving anyway!

                logger.info('(%(i)d/%(count_here)d) move %(to_move)r ...', locals())
                cp = from_o.manage_cutObjects(ids=(to_move,))
                to_o.manage_pasteObjects(cp)
                transaction.commit()
                cnt['moved_total'] += 1
                if move_limit and cnt['moved_total'] >= move_limit:
                    logger.info('Total move limit exceeded (%(move_limit)r)', locals())
                    return i
            return i

        local_limits = []
        if move_limit_each is not None:
            local_limits.append(move_limit_each)
        if move_limit is not None:
            local_limits.append(move_limit - cnt['moved_total'])
        unlimited = not local_limits
        if not unlimited:
            local_limit = min(local_limits)
            if local_limit >= count_here:
                unlimited = True

        if unlimited:
            batch_size = pop('batch_size', 10)
            all_ids = [brain.getId for brain in brains]
            # TODO: use visaplan.tools.batches.batch_tuples
            if batch_size and batch_size < count_here:
                logger.info('Moving %(count_here)d %(portal_type)r objects'
                ' of language %(lang)r'
                ' in batches of %(batch_size)d'
                ' to %(to_o)r ...', locals())
                total_found = len(all_ids)
                full_batches, in_last = divmod(total_found, batch_size)
                total_batches = full_batches
                if in_last:
                    total_batches += 1
                batch_nr = 0
                while all_ids:
                    batch = all_ids[:batch_size]
                    del     all_ids[:batch_size]
                    first_nr = batch_nr * batch_size + 1
                    last_nr = min(total_found, (batch_nr+1) * batch_size)
                    this_batch_size = last_nr + 1 - first_nr
                    batch_nr += 1
                    logger.info('Objects %(first_nr)d to %(last_nr)d'
                                ' (batch %(batch_nr)d / %(total_batches)d) ...',
                                locals())
                    cp = from_o.manage_cutObjects(ids=batch)
                    res = to_o.manage_pasteObjects(cp)
                    transaction.commit()
            else:
                logger.info('Moving %(count_here)d %(portal_type)r objects of language %(lang)r to %(to_o)r ...', locals())
                cp = from_o.manage_cutObjects(ids=all_ids)
                res = to_o.manage_pasteObjects(cp)
                transaction.commit()
                pp(res=res)
            logger.info('Done: Pasted %(count_here)d objects in %(to_o)r', locals())
            cnt['moved_total'] += count_here
            return count_here

        ids = []
        i = 0
        for brain in brains:
            ids.append(brain.getId)
            i += 1
            if i >= local_limit:
                break

        logger.info('Moving %(local_limit)d / %(count_here)d %(portal_type)r objects of language %(lang)r to %(to_o)r ...', locals())
        cp = from_o.manage_cutObjects(ids=ids)
        to_o.manage_pasteObjects(cp)
        logger.info('Done: Pasted %(local_limit)d objects in %(to_o)r', locals())
        cnt['moved_total'] += local_limit
        return local_limit
    finally:
        transaction.commit()
    # ----------------------------------------------- ] ... _move_objects ]

def _skip_language(la, dic):
    """
    Helper for _clone_tree_inner: Skip the given language?
    """
    logger = dic['logger']
    allowed_languages = dic.get('allowed_languages')
    if not allowed_languages:
        logger.warn('no allowed languages information')
        return False
    if la in allowed_languages:
        return False




# --------------------------------- [ userinfo_factory ... [
def userinfo_factory(context, pretty=0, forlist=0,
                     title_or_id=0):
    """
    Wie groupinfo_factory, aber für Benutzerobjekte.

    pretty -- für Benutzer: title als formatierter Name
              (author.get_formatted_name), wenn möglich
    forlist -- Minimale Rückgabe (nur ID und Titel),
               aber mit pretty kombinierbar
    title_or_id -- für Verwendung mit visaplan.tools.classes.Proxy:
               gib nur den Title oder ersatzweise die ID zurück
               (mit pretty kombinierbar)
    """
    acl = getToolByName(context, 'acl_users')
    acl_gu = acl.getUser
    if pretty or not forlist:
        author = context.restrictedTraverse('@@author', None)
        gbbuid = author.getBrainByUserId
        gfn = author.get_formatted_name

    # ---------------------------- [ forlist ... [
    def basic_user_info(member_id):
        """
        Basisinformationen über einen Benutzer:
        id, title

        forlist, not: pretty
        """
        member = acl_gu(member_id)
        if member:
            return {'id': member_id,
                    'title': member.getProperty('fullname'),
                    }

    def pretty_user_info(member_id):
        """
        Basisinformationen über einen Benutzer:
        id, title

        forlist, pretty
        """
        member = acl_gu(member_id)
        if member:
            brain = gbbuid(member_id)
            return {'id': member_id,
                    'title': (brain and gfn(brain))
                             or member.getProperty('fullname'),
                    }
    # ---------------------------- ] ... forlist ]

    # ------------------------ [ title_or_id ... [
    def pretty_title_or_id(member_id):
        try:
            return pretty_user_info(member_id)['title'] \
                   or member_id
        except:
            return None

    def basic_title_or_id(member_id):
        try:
            return basic_user_info(member_id)['title'] \
                   or member_id
        except:
            return None
    # ------------------------ ] ... title_or_id ]

    def full_user_info(member_id):
        """
        not: forlist, not: pretty
        """
        member = acl_gu(member_id)
        if member:
            brain = gbbuid(member_id)
            return {'id': member_id,
                    'title': member.getProperty('fullname'),
                    'brain': brain,
                    'email': member.getProperty('email'),
                    }

    def full_pretty_user_info(member_id):
        """
        not: forlist, pretty
        """
        member = acl_gu(member_id)
        if member:
            brain = gbbuid(member_id)
            return {'id': member_id,
                    'title': (brain and gfn(brain))
                             or member.getProperty('fullname'),
                    'brain': brain,
                    'email': member.getProperty('email'),
                    }

    if title_or_id:
        if pretty:
            return pretty_title_or_id
        else:
            return basic_title_or_id
    if forlist:
        if pretty:
            return pretty_user_info
        else:
            return basic_user_info
    if pretty:
        return full_pretty_user_info
    else:
        return full_user_info
# --------------------------------- ] ... userinfo_factory ]



def is_direct_member__factory(context, userid):
    """
    Erzeuge eine Funktion, die prüft, ob der *beim Erzeugen angegebene* Nutzer
    in der bei jedem Aufruf anzugebenden Gruppe direkt enthalten ist.
    """
    acl = getToolByName(context, 'acl_users')
    gpm = acl.source_groups._group_principal_map

    def is_direct_member_of(group_id):
        return userid in gpm[group_id]

    return is_direct_member_of


def is_member_of__factory(context, userid):
    """
    Erzeuge eine Funktion, die die *direkte oder indirekte* Mitgliedschaft
    des übergebenen Users in der jeweils zu übergebenden Gruppe überprüft.
    """
    acl = getToolByName(context, 'acl_users')
    gpm = acl.source_groups._group_principal_map

    groups = build_groups_set(gpm, userid)

    def is_member_of(groupid):
        return groupid in groups

    return is_member_of


def is_member_of_any(context, group_ids, user_id=None, default=False):
    """
    Is the given or active user member of one of the given groups?

    - For anonymous execution, raises Unauthorized
      (might become changable by keyword-only argument)
    - The normal usage is without specification of the user_id,
      i.e. checking for the logged-in user
    - if the group_ids sequence is empty, the default is used

    """
    pm = getToolByName(context, 'portal_membership')
    if pm.isAnonymousUser():
        raise Unauthorized
    if user_id is not None:
        # wer darf sowas fragen? Manager, Gruppenmanager, ...?
        # TODO: Hier entsprechende Überprüfung!
        pass

    if not group_ids:
        return default
    if user_id is None:
        member = pm.getAuthenticatedMember()
        user_id = member.getId()

    return user_id in get_all_members(context, group_ids)


def get_all_members(context, group_ids, **kwargs):  # --- [[
    """
    Return all members of the given groups

    Liefere alle Mitglieder der übergebenen Gruppe(n).

    Schlüsselwortargumente für .utils.recursive_members und
    groupinfo_factory dürfen angegeben werden;
    letztere werden aber ignoriert, wenn die vorgabegemäße Filterung
    groups_only=True übersteuert wird. In diesem Fall (potentiell sowohl
    Benutzer als auch Gruppen, oder nur Benutzer) werden nur IDs
    zurückgegeben.

    Rückgabe:
    - Sequenz von Gruppeninformationen, mit groups_only=True (Vorgabe);
    - ansonsten nur eine Sequenz der IDs (je nach Aufrufargumenten nur
      Benutzer-IDs, oder gemischt)
    """
    acl = getToolByName(context, 'acl_users')
    gpm = acl.source_groups._group_principal_map
    filter_args = {}
    for key in ('groups_only', 'users_only',
		'containers',
		'default_to_all'):
	try:
	    filter_args[key] = kwargs.pop(key)
	except KeyError:
	    pass
    members = recursive_members(list_of_strings(group_ids),
				gpm, **filter_args)
    groups_only = filter_args.get('groups_only', False)
    if groups_only:
	format_args = {'pretty': False,
		       'forlist': True,
		       }
	format_args.update(kwargs)
	ggibi = groupinfo_factory(context, **format_args)
	res = []
	for gid in members:
	    res.append(ggibi(gid))
	return res
    elif kwargs and debug_active:
	pp('ignoriere:', kwargs)

    return members



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Helpers for arguments exploitation
"""

# Python compatibility:
from __future__ import absolute_import, print_function

try:
    # visaplan:
    from visaplan.tools.minifuncs import check_kwargs
except ImportError as e:
    if __name__ == '__main__':
        def dummy(*args, **kwargs):
            pass
        check_kwargs = dummy
        print('W: %(e)r Some tests might fail' % locals())
    else:
        raise

__all__ = [
        # Helferlein für **kwargs:
        'extract_object_and_brain', # --> (o, brain)
        'extract_object_or_brain',  # --> (o, brain=None)
        'extract_brain_or_object',  # --> (brain, o=None)
        # [switch_]menu:
        'extract_menu_switch',
        'normalize_menu_switch',
        'extract_layout_switch',
        # query manipulation:
        'apply_move_order_options',
        ]


def extract_object_or_brain(dic, do_pop=True):
    """
    Extrahiere die Schlüssel 'o' (ein Objekt) und 'brain' aus dem übergebenen
    Dict; gib ein 2-Tupel (o, brain) zurück.

    Der Zweck ist die Ermittlung des Objekts; wenn 'brain' nicht enthalten ist,
    wird hierfür None zurückgegeben.

    dic -- das Python-Dict-Objekt
    do_pop -- wenn True (Vorgabe), werden die Schlüssel aus dem Dict entfernt

    >>> dic1 = {'brain': MockBrain(), 'other': 1}
    >>> extract_object_or_brain(dic1)
    (<an object>, <a brain>)
    >>> dic1
    {'other': 1}
    """
    if do_pop:
        get = dic.pop
    else:
        get = dic.get
    if 'brain' in dic:
        brain = get('brain')
        if 'o' in dic:
            o = get('o', None)
        else:
            o = None
        if o is None:
            o = brain.getObject()
    else:
        o = get('o')
        brain = None
    return o, brain


def extract_object_and_brain(dic, do_pop=True):
    """
    Extrahiere die Schlüssel 'o' (ein Objekt) und 'brain' aus dem übergebenen
    Dict; gib ein 2-Tupel (o, brain) zurück.

    Das jeweils fehlende wird ermittelt; eines von beiden muß natürlich
    enthalten sein.

    dic -- das Python-Dict-Objekt
    do_pop -- wenn True (Vorgabe), werden die Schlüssel aus dem Dict entfernt

    >>> dic11 = {'brain': MockBrain(), 'other': 11}
    >>> extract_object_and_brain(dic11)
    (<an object>, <a brain>)
    >>> dic11
    {'other': 11}
    >>> dic12 = {'o': MockObject(), 'other': 12}
    >>> extract_object_and_brain(dic12)
    (<an object>, <a brain>)
    """
    if do_pop:
        get = dic.pop
    else:
        get = dic.get
    if 'brain' in dic:
        brain = get('brain')
        if 'o' in dic:
            o = get('o', None)
        else:
            o = None
        if o is None:
            o = brain.getObject()
    else:
        o = get('o')
        brain = None
    if brain is None:
        brain = o.getHereAsBrain()
    return o, brain


def extract_brain_or_object(dic, do_pop=True):
    """
    Extrahiere die Schlüssel 'o' (ein Objekt) und 'brain' aus dem übergebenen
    Dict; gib ein 2-Tupel (brain, o) zurück.

    Der Zweck ist die Ermittlung des 'Brains'; wenn 'o' nicht enthalten ist,
    wird hierfür None zurückgegeben.

    dic -- das Python-Dict-Objekt
    do_pop -- wenn True (Vorgabe), werden die Schlüssel aus dem Dict entfernt

    Wenn das 'brain' schon enthalten ist, braucht das Objekt nicht beschafft zu
    werden:
    >>> dic21 = {'brain': MockBrain(), 'other': 21}
    >>> extract_brain_or_object(dic21)
    (<a brain>, None)
    >>> dic21
    {'other': 21}

    >>> dic22 = {'o': MockObject(), 'other': 22}
    >>> extract_brain_or_object(dic22)
    (<a brain>, <an object>)
    """
    if do_pop:
        get = dic.pop
    else:
        get = dic.get
    if 'brain' in dic:
        brain = get('brain')
        if 'o' in dic:
            o = get('o', None)
        else:
            o = None
    else:
        o = get('o')
        brain = None
    if brain is None:
        brain = o.getHereAsBrain()
    return brain, o


def extract_menu_switch(dic, created, default=None, do_pop=True):
    """
    Extract the switch_menu and/or menu (boolean) value
    from the given dict; if both are present but different,
    a ValueError is raised.

    Other than normalize_menu_switch (below), this function takes other
    keys into account as well, e.g. "created".

    We'll use a little test helper which enforces the `created` argument to be
    given by name, for enhanced clarity:
    >>> def ems(dic, **kwargs):
    ...     return extract_menu_switch(dic, **kwargs)

    >>> dic={'switch_menu': 1, 'menu': 1, 'other': 42}
    >>> sorted(dic.items())
    [('menu', 1), ('other', 42), ('switch_menu', 1)]
    >>> ems(dic, created=None, do_pop=False)
    1

    With do_pop=False, the dictionary is not changed:
    >>> sorted(dic.items())
    [('menu', 1), ('other', 42), ('switch_menu', 1)]

    The default is, however, to remove the used options:
    >>> ems(dic, created=None)
    1
    >>> dic
    {'other': 42}

    1 and True are equivalent:
    >>> dic2={'switch_menu': 1, 'menu': True, 'other': 43}
    >>> extract_menu_switch(dic2, created=None)
    1
    >>> dic2
    {'other': 43}

    If both menu and switch_menu are given but different, we'll get an error:
    >>> ems({'switch_menu': 1, 'menu': 0}, created=None)
    Traceback (most recent call last):
    ...
    ValueError: 'menu': 0 conflicts 'switch_menu' value 1!

    With created=True, the default is to *not* activate the menu entry:
    >>> ems({}, created=True)
    False

    With set_menu=False, None will be returned:
    >>> ems({'switch_menu': 1, 'menu': True, 'set_menu': False}, created=None)

    Without a 'set_menu' value, the default depends on the `created` argument.
    For freshly created items, the default is "no change":
    >>> ems({'switch_menu': 1}, created=True)

    >>> ems({'switch_menu': 1}, created=False)
    1

    """
    if do_pop:
        get = dic.pop
    else:
        get = dic.get
    res = None
    keys = []
    for key in (
        'switch_menu', 'menu',
        ):
        val = get(key, None)
        if val is not None:
            if res is None:
                res = val
            elif val != res:
                earlier = keys[1:] and tuple(keys) or keys[0]
                raise ValueError('%(key)r: %(val)r conflicts '
                                 '%(earlier)r value %(res)r!'
                                 % locals())
            keys.append(key)
    set_menu = get('set_menu', None)
    if set_menu is None:
        if created:
            if res is None:
                return False
            else:
                return None
        return res
    elif not set_menu:
        return None
    elif res is None:
        return default
    else:
        return res


def normalize_menu_switch(dic):
    """
    Return the value of the first menu-switching option,
    check for conflicts, and remove superfluous (but not conflicting)
    keys.

    >>> dic1={'switch_menu': 1, 'menu': 1, 'other': 42}
    >>> normalize_menu_switch(dic1)
    1
    >>> sorted(dic1.items())
    [('other', 42), ('switch_menu', 1)]

    The keys are not added if missing:
    >>> dic2={'other': 43}
    >>> normalize_menu_switch(dic2)
    >>> sorted(dic2.items())
    [('other', 43)]

    Conflicting specifications cause an error:
    >>> dic3={'switch_menu': 1, 'menu': 0}
    >>> normalize_menu_switch(dic3)
    Traceback (most recent call last):
      ...
    ValueError: 'menu': 0 conflicts 'switch_menu' value 1!

    The function is sometimes used just to normalize the given dict,
    ignoring the return value:

    >>> dic1={'switch_menu': 1, 'menu': 1, 'other': 42}
    >>> def f(dic):
    ...     normalize_menu_switch(dic)
    ...     return sorted(dic.items())
    >>> f(dic1)
    [('other', 42), ('switch_menu', 1)]

    The keys are not added if missing:
    >>> f({'other': 43})
    [('other', 43)]

    """
    get = dic.get
    res = None
    keys = []
    allowed = (
        'switch_menu',
        'menu',
        )
    first = True
    firstval = None
    for key in allowed:
        if first:
            val = dic.get(key, None)
            firstval = val
            first = False
        else:
            val = dic.pop(key, None)
        if val is not None:
            if res is None:
                res = val
            elif val != res:
                earlier = keys[1:] and tuple(keys) or keys[0]
                raise ValueError('%(key)r: %(val)r conflicts '
                                 '%(earlier)r value %(res)r!'
                                 % locals())
            keys.append(key)
    if res is not None and firstval is None:
        dic[allowed[0]] = res
    return res


def extract_layout_switch(dic, default=None, do_pop=True):
    """
    Extract the layout and/or view_id value
    from the given dict; if both are present but different,
    a ValueError is raised.

    Futhermore, the "layout" might be got (.getLayout) or set
    (.selectViewTemplate);
    thus, a 3-tuple ('layout_id', do_set, do_get) is returned.
    """
    if do_pop:
        get = dic.pop
    else:
        get = dic.get
    layout = None
    layout_given = False
    set_given = 'set_layout' in dic
    set_layout = get('set_layout', None)
    get_given = 'get_layout' in dic
    get_layout = get('get_layout', None)
    keys = []
    for key in (
        'layout', 'view_id',
        ):
        val = get(key, None)
        if val is not None:
            if layout is None:
                layout = val
            elif val != layout:
                earlier = keys[1:] and tuple(keys) or keys[0]
                raise ValueError('%(key)r: %(val)r conflicts '
                                 '%(earlier)r value %(layout)r!'
                                 % locals())
            keys.append(key)
            layout_given = True
    if layout is None:
        layout = default

    if set_layout is None:
        set_layout = bool(layout)

    return (layout, set_layout, get_layout)


def _extract_move_args(dic, default=None, do_pop=True):
    """
    Only for _tree.clone_tree for now, and not yet generalized;
    might become replaced, removed or renamed!

    Take a (usually: **kwargs) dict and extract (by default: pop)
    the keys move_children, move_limit and move_limit_each

    Return a 3-tuple (move_children, move_limit, move_limit_each);
    move_children defaults to True, if at least one of the other keys
    were contained.

    Arguments:

    dic - the dictionary
    default - default value for move_limit* keys
    do_pop - whether to pop the keys from the dict.

    >>> dic = {'move_limit': 1, 'other_option': 42}
    >>> _extract_move_args(dic)
    (True, 1, None)
    >>> dic
    {'other_option': 42}
    >>> dic = {'other_option': 42}
    >>> _extract_move_args(dic, 3)
    (False, 3, 3)
    """
    if do_pop:
        get = dic.pop
    else:
        get = dic.get
    res = [None]
    has_other = False

    for otherkey in ('move_limit', 'move_limit_each'):
        if otherkey in dic:
            val = get(otherkey)
            has_other = True
        else:
            val = get(otherkey, default)
        res.append(val)
    res[0] = get('move_children', has_other)
    return tuple(res)


def setdefault_move_args(dic, default=None):
    """
    Like _extract_move_args (above), but never pops keys from the dict
    but rather amends the missing keys (since we seem to work recursively now
    and re-use the kwargs dictionaries down the trip).

    Since the {}.setdefault method returns the value set, and for testability,
    it returns the same 3-tuple like _extract_move_args.

    >>> dic = {'move_limit': 1, 'other_option': 42}
    >>> setdefault_move_args(dic)
    (True, 1, None)

    >>> def si(dic): return sorted(dic.items())
    >>> si(dic)
    [('move_children', True), ('move_limit', 1), ('move_limit_each', None), ('other_option', 42)]
    >>> dic = {'other_option': 42}
    >>> setdefault_move_args(dic, 3)
    (False, 3, 3)
    >>> si(dic)
    [('move_children', False), ('move_limit', 3), ('move_limit_each', 3), ('other_option', 42)]
    """
    res = [None]
    has_other = False
    get = dic.get

    for otherkey in ('move_limit', 'move_limit_each'):
        if otherkey in dic:
            val = get(otherkey)
            has_other = True
        else:
            dic[otherkey] = default
            val = default
        res.append(val)
    res[0] = dic.setdefault('move_children', has_other)
    return tuple(res)


def setdefault_source_language(dic, default=None):
    """
    This function modifies the given dict in-place;
    like the {}.setdefault method, it returns the actual value.

    Besides the verbose 'source_language' key, the 'src_lang' alias
    is considered as well. If both are found, they must be equal.

    >>> dic={'src_lang': 'de'}
    >>> setdefault_source_language(dic, None)
    'de'
    >>> dic
    {'source_language': 'de'}
    >>> dic={'src_lang': 'de', 'source_language': 'en'}
    >>> setdefault_source_language(dic, None)
    Traceback (most recent call last):
    ...
    ValueError: Conflicting values for source_language ('en') and src_lang ('de')!
    """
    res = []
    has_other = False
    first = True

    for otherkey in ('source_language', 'src_lang'):
        if first:
            get = dic.get
            first = False
        else:
            get = dic.pop
        if otherkey in dic:
            res.append(get(otherkey))
    if len(set(res)) == 1:
        dic['source_language'] = res[0]
        return res[0]
    elif not res:
        dic['source_language'] = default
        return default
    else:
        raise ValueError('Conflicting values for source_language (%r)'
                ' and src_lang (%r)!' % tuple(res))


def apply_move_order_options(query, kwargs, **kw):
    """
    This function modifies the given `query` in-place, by default
    (`do_pop`=True) consuming the given kwargs dictionary.

    Since we build a query for move operations, we like to maintain a given
    position in the source folder.

    We'll use two little test helpers here:

    >>> def amoo(query, **kwargs):  # finalizing (default)
    ...     apply_move_order_options(query, kwargs)
    ...     return sorted(query.items())
    >>> def amoo_nf(query, **kwargs):  # non-finalizing
    ...     apply_move_order_options(query, kwargs, finalize=False)
    ...     return sorted(query.items())

    If a depth of 1 is specified, the default order is by position in parent:
    >>> amoo_nf({'path': '/Plone/some/path', 'depth': 1})
    [('depth', 1), ('path', '/Plone/some/path'), ('sort_on', 'getObjPositionInParent')]
    >>> amoo_nf({'path': '/Plone/some/path'}, depth=1)
    [('depth', 1), ('path', '/Plone/some/path'), ('sort_on', 'getObjPositionInParent')]

    The default ordering by position in parent is fine for moving contents from
    a single source folder to a single destination; but if you are collecting
    objects from several folders, most likely this would either yield
    unpredictable results, or you'd reflect the order of the parents processed
    in the result.  In such cases, you'll want to sort by some more reasonable
    value, e.g. by some date. The default date index used is 'created':

    >>> q2 = {'path': '/Plone/some/path'}
    >>> amoo_nf(q2)
    [('path', '/Plone/some/path'), ('sort_on', 'created')]

    ... but you can specify another date which is is used unless a depth of 1
    is given (or some other sorting specification is already present):

    >>> q3 = {'path': '/Plone/some/path'}
    >>> amoo_nf(q3, use_date='effective')
    [('path', '/Plone/some/path'), ('sort_on', 'effective')]

    >>> q4 = {'path': '/Plone/some/path', 'depth': 1}
    >>> amoo_nf(q4, use_date='effective')
    [('depth', 1), ('path', '/Plone/some/path'), ('sort_on', 'getObjPositionInParent')]

    So far, we have suppressed finalization, because this yields simpler
    results.
    To make sure you get a usable query dict, specify `finalize=True`:
    >>> amoo(q4)
    [('path', {'operator': 'or', 'query': ['/Plone/some/path'], 'depth': 1}), ('sort_on', 'getObjPositionInParent')]

    >>> q5 = {'path': '/Plone/some/path'}
    >>> amoo_nf(q5, sort_on='effective')
    [('path', '/Plone/some/path'), ('sort_on', 'effective')]
    >>> amoo(q5)
    [('path', {'operator': 'or', 'query': ['/Plone/some/path']}), ('sort_on', 'effective')]

    We can do this in one step as well, of course:

    >>> q6 = {'path': '/Plone/some/path'}
    >>> amoo(q6, sort_on='effective')
    [('path', {'operator': 'or', 'query': ['/Plone/some/path']}), ('sort_on', 'effective')]

    The depth specification is supposed to live in the `path` specification.
    With finalize=True, this will be rectified.

    Finally, here are the supported keyword-only options:

    `finalize` (default: True), create a *valid* `path` filter (if given)
    `do_pop` (default: True), pop the keys from the "kwargs" which contain the
             defaults
    `strict` (default: True), raise an error if undefined keyword arguments are
             specified (see the `check_kwargs` function)

    """
    do_pop = kw.pop('do_pop', True)
    finalize = kw.pop('finalize', True)
    check_kwargs(kw)
    if do_pop:
        get = kwargs.pop
    else:
        get = kwargs.get

    path_given = 'path' in query
    depth_given = 'depth' in query
    depth_root = query.get('depth')
    depth = None
    if path_given:
        path = query.get('path')
        if isinstance(path, dict):
            depth = path.get('query')
            if (depth      is not None and
                depth_root is not None and
                depth != depth_root):
                raise ValueError('conflicting depth specs! '
                                 ' (depth=%(depth_root)r, '
                                  ' path=%(path)s)'
                                  % locals())
    if (depth      is     None and
        depth_root is not None):
        depth = depth_root
    if depth is None:
        depth = get('depth', None)
    if depth is not None:
        query['depth'] = depth
    if 'sort_on' not in query:
        if 'sort_on' in kwargs:
            sort_on = get('sort_on')
            if sort_on:
                query['sort_on'] = sort_on
        elif depth == 1:
            query['sort_on'] = 'getObjPositionInParent'
        elif depth != 0:
            use_date = get('use_date', None) or 'created'
            query['sort_on'] = use_date
    if finalize:
        if 'path' in query:
            path = query['path']
            if isinstance(path, dict):
                path_dict = path
            elif isinstance(path, list):
                path_dict = {
                    'query': path,
                    'operator': 'or',
                    }
            elif isinstance(path, tuple):
                path_dict = {
                    'query': list(path),
                    'operator': 'or',
                    }
            else:
                path_dict = {
                    'query': path.split(),
                    'operator': 'or',
                    }
            if 'depth' in query:
                del query['depth']
            if depth is not None:
                path_dict['depth'] = depth
            query['path'] = path_dict


if __name__ == '__main__':
    # added to ../mock.py as well:
    class MockBrain(object):
      def __repr__(self):
        return '<a brain>'
      def getObject(self):
        return MockObject()
    class MockObject(object):
      def __repr__(self):
        return '<an object>'
      def getHereAsBrain(self):
        return MockBrain()
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: tw=79 cc=+1 sw=4 sts=4 et si
"""
visaplan.plone.tools: attools: Archetypes-related tools (_at6)
"""

def can_be_html(field):
    # for use as a predicate for schema.filterFields
    # (see the Products.Archetypes.Schema.Schemata class):
    # for a field to match,
    # each predicate function must return something truish
    act = getattr(field, 'allowable_content_types', None)
    if act is None:
        return 0
    if not isinstance(act, (list, tuple)):
        return 0
    return 'text/html' in act




# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
visaplan.plone.tools.search - search helpers
"""
# Python compatibility:
from __future__ import absolute_import, print_function

from six import text_type as six_text_type

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"
VERSION = (0,
           3,   # make_querystring_mangle()
           )

__all__ = [
        'normalizeQueryString',  # created by factory:
        'make_querystring_normalizer',
        'mangle_umlauts',
        'mangleQueryString',     # created by factory:
        'make_querystring_mangle',
        'language_spec',
        ]

try:
    # Zope:
    from Products.CMFCore.utils import getToolByName
except ImportError:
    if __name__ == '__main__':
        class MockLanguageTool(object):
            def listSupportedLanguages():
                return [('en', 'English'), ('de', 'Deutsch')]
        def getToolByName(context, name):
            if name == 'portal_languages':
                return MockLanguageTool
    else:
        raise

# visaplan:
from visaplan.tools.lands0 import list_of_strings

try:
    # visaplan:
    from visaplan.tools.coding import safe_decode
except (ImportError, ValueError):
    if __name__ == '__main__':
        safe_decode = six_text_type
        print('Install visaplan.tools to properly test the normalizeQueryString function!')
    else:
        raise


def make_querystring_normalizer(decode=safe_decode):
    r"""
    Factory to create a `normalizeQueryString` function,
    injecting the default decoding function;
    by default, the safe_decode function from visaplan.tools is used,
    which (of course) recognizes unicode strings and decodes UTF8 and Latin-1.

    You may want to replace this `decode` argument by your own function,
    e.g. to

    - raise an exception if a non-unicode string is given,
      or
    - log calls which use non-unicode strings (to help during development)

    This factory is called by the module code to create the default
    `normalizeQueryString` function:

    >>> normalizeQueryString = make_querystring_normalizer(safe_decode)

    ... which will accept a string (and optionally another decoding function)
    and return a (possibly empty) list of unicode strings.

    For empty input, an empty list is returned:

    >>> normalizeQueryString('')
    []
    >>> normalizeQueryString(u'')
    []
    >>> normalizeQueryString('   \t \n \r')
    []

    There is special handling for the asterisk character ("*"):

    - If it is not present anywhere in the (non-empty) input string,
      it is added to the beginning and end of every string:

      >>> normalizeQueryString('abc')
      [u'abc*']
      >>> normalizeQueryString(u'abc')
      [u'abc*']
      >>> normalizeQueryString('   abc\t  def  ')
      [u'abc*', u'def*']
      >>> normalizeQueryString('entkoppelte Förderschnecke ')
      [u'entkoppelte*', u'F\xf6rderschnecke*']

    - If it *is* contained somewhere, we assume the user to do this on purpose,
      and thus don't amend the asterisk anywhere:

      >>> normalizeQueryString('ab*c')
      [u'ab*c']
      >>> normalizeQueryString('rohr* leitung  ')
      [u'rohr*', u'leitung']

    - a single-word asterisk is removed and simply causes the asterisk
      amemdments to be suppressed:

      >>> normalizeQueryString('rohr * leitung  ')
      [u'rohr', u'leitung']

    - Of course, a single asterisk *only* is like no search expression at all:

      >>> normalizeQueryString('  * ')
      []
    """
    def normalizeQueryString(string, decode=decode):
        """
        Return a (possibly empty) list of unicode strings.

        Arguments:

          string -- a string of some kind

          decode -- a function to check and, if necessary,
                    decode the given string
        """
        # searchstring is None or empty
        if not string:
            return []

        if not isinstance(string, six_text_type):
            string = decode(string)
        if string.strip() == u'*':
            return []
        has_asterisk = u'*' in string
        strings = string.split()
        if has_asterisk or u'*' in strings:
            # ein alleinstehendes Asterisk würde *immer* passen;
            # also nur zur Unterdrückung der Automatik nutzen:
            return [s for s in strings
                    if s != u'*'
                    ]
        else:
            return [s + u'*'
                    for s in strings
                    ]

    return normalizeQueryString
normalizeQueryString = make_querystring_normalizer(safe_decode)

UMLAUTS_TO_BASECHAR = {
    u'Ä': u'A',
    u'Ö': u'O',
    u'Ü': u'U',
    u'ä': u'a',
    u'ö': u'o',
    u'ü': u'u',
    u'ß': u'ss',
    }
def mangle_umlauts(s):
    ur"""
    We mimic a conversion here which we found done to our catalog data.

    >>> mangle_umlauts(u'r\xfcckstellkr\xe4fte')
    u'ruckstellkrafte'
    >>> mangle_umlauts(u'sch\xf6ne T\xf6ne')
    u'schone Tone'
    """
    nonascii = set(s)
    for ch in sorted(nonascii.intersection(UMLAUTS_TO_BASECHAR.keys())):
        s = s.replace(ch, UMLAUTS_TO_BASECHAR[ch])
    return s


def make_querystring_mangle(decode=safe_decode, *funcs):
    r"""
    Factory to create a `mangleQueryString` function,
    injecting the default decoding function;
    by default, the safe_decode function from visaplan.tools is used,
    which (of course) recognizes unicode strings and decodes UTF-8 and Latin-1.

    This extends the normalizeQueryString functionality by applying one or more
    transformations which are most interesting when non-ASCII search
    expressions join the game.

    >>> mangleQueryString = make_querystring_mangle(safe_decode)

    ... which will accept a string (and optionally another decoding function)
    and return a (possibly empty) list of unicode strings.

    For empty input, an empty list is returned:

    >>> mangleQueryString('   \t \n \r')
    []

    Like with normalizeQueryString,
    there is special handling for the asterisk character ("*");
    but additionally, we transform to lower case:

    - If it is not present anywhere in the (non-empty) input string,
      it is added to the beginning and end of every string:

      >>> mangleQueryString('Abc')
      [u'abc*']
      >>> mangleQueryString(u'Abc')
      [u'abc*']
      >>> mangleQueryString('   aBc\t  def  ')
      [u'abc*', u'def*']
      >>> mangleQueryString('entkoppelte Förderschnecke ')
      [u'entkoppelte*', u'f\xf6rderschnecke*']

    - If it *is* contained somewhere, we assume the user to do this on purpose,
      and thus don't amend the asterisk anywhere:

      >>> mangleQueryString('ab*c')
      [u'ab*c']
      >>> mangleQueryString('rohr* leitung  ')
      [u'rohr*', u'leitung']

    - a single-word asterisk is removed and simply causes the asterisk
      amemdments to be suppressed:

      >>> mangleQueryString('Rohr * Leitung  ')
      [u'rohr', u'leitung']

    - Of course, a single asterisk *only* is like no search expression at all:

      >>> mangleQueryString('  * ')
      []

    Now, so far the only difference to normalizeQueryString was the lowercase
    conversion; not very interesting. But we can add more conversions:

      >>> mqs = make_querystring_mangle(None, mangle_umlauts)
      >>> mqs(u'entkoppelte F\xf6rderschnecke ')
      [u'entkoppelte*', u'f\xf6rderschnecke*', u'forderschnecke*']
      >>> mqs(u'R\xfcckstellkr\xe4fte *')
      [u'r\xfcckstellkr\xe4fte', u'ruckstellkrafte']
      >>> mqs(u'R\xfcckstellkr\xe4fte')
      [u'r\xfcckstellkr\xe4fte*', u'ruckstellkrafte*']

    """
    if decode is None:
        decode = safe_decode
    FUNCS = tuple(funcs)

    def mangleQueryString(string, decode=decode):
        """
        Return a (possibly empty) list of unicode strings.

        Arguments:

          string -- a string of some kind

          decode -- a function to check and, if necessary,
                    decode the given string
        """
        # searchstring is None or empty
        if not string:
            return []

        if not isinstance(string, six_text_type):
            string = decode(string)
        if string.strip() == u'*':
            return []
        has_asterisk = u'*' in string
        res = []
        for chunk in string.lower().split():
            if chunk == u'*':
                assert has_asterisk
                continue
            if chunk in res:
                continue
            res.append(chunk)
            for f in FUNCS:
                val = f(chunk)
                if val not in res:
                    res.append(val)
        if has_asterisk:
            return res
        else:
            return [s + u'*'
                    for s in res
                    ]

    return mangleQueryString
mangleQueryString = make_querystring_mangle(safe_decode, mangle_umlauts)


def language_spec(value=None, form=None, context=None,
                  language_tool=None,
                  default_to_all=True,
                  all_languages=None):
    r"""
    Gib den korrekten Suchausdruck für die Einschränkung auf bestimmte
    Sprachversionen zurück.

    Argumente (bitte sämtlich benannt angeben!):

    value -- der Wert, falls schon ermittelt (None, String oder Sequenz)
    form -- das form-dict, falls nicht <value> angegeben; ggf. aus dem
            <context> ermittelt
    context -- kann ggf. verwendet werden, um das <language_tool> und/oder
               <form> zu ermitteln
    all_languages -- der Wert, der im Falle des Vorkommens von "all" im <value>
                     hinzugefügt wird; Unterdrücken z. B. durch Angabe von []
    default_to_all -- soll bei leerer oder fehlender Angabe eine
                      sprachunabhängige Suche ausgeführt werden?
                      (zu überlegen ist dann, ob *überhaupt* eine Angabe
                      benötigt wird.  Vielleicht ja, wegen etwaiger
                      Default-Mechanismen in aufgerufenen Funktionalitäten ...)

    >>> kwargs={'all_languages': ['de', 'en']}
    >>> language_spec(value='', **kwargs)
    ['de', 'en']
    >>> language_spec(value='all', **kwargs)
    ['de', 'en']
    >>> language_spec(value=['all', 'fr'], **kwargs)
    ['de', 'en', 'fr']

    'all' wird auch mit Leerraum erkannt:
    >>> language_spec(value=' all\n', **kwargs)
    ['de', 'en']

    Wenn der <value> nicht übergeben wird, wird er der <form>-Variablen
    'language' entnommen:

    >>> form={'language': 'de'}
    >>> kwargs.update({'form': form})
    >>> language_spec(**kwargs)
    ['de']

    Um ggf. eine leere Liste zu erhalten, muß default_to_all=False angegeben
    werden:

    >>> kwargs.update({'default_to_all': False})
    >>> language_spec(**kwargs)
    ['de']
    >>> language_spec(value='', **kwargs)
    []
    """
    if value is None:
        if form is None:
            form = context.REQUEST.form
        try:
            value = form['language']
        except KeyError:
            value = []
    values = set(list_of_strings(value))
    try:
        values.remove('all')
    except KeyError:
        if values:
            has_all = False
        else:
            has_all = default_to_all
    else:
        has_all = True
    if has_all:
        if all_languages is None:
            if language_tool is None:
                language_tool = getToolByName(context, 'portal_languages')
            all_languages = [tup[0]
                             for tup in language_tool.listSupportedLanguages()
                             ]
        values.update(all_languages)
    return sorted(values)


if __name__ == "__main__":
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
visaplan.plone.tools._at0 -- imported via .attools
"""

# Python compatibility:
from __future__ import absolute_import

# Local imports:
from visaplan.plone.tools._have import HAS_ARCHETYPES

if HAS_ARCHETYPES:
    # Zope:
    from Products.Archetypes.utils import mapply, shasattr


# <http://stackoverflow.com/a/35756881/1051649>;
# von maurits, <http://stackoverflow.com/users/621201/maurits>:
def initialize_rich_text_fields(instance):
    """New rich text fields should have mimetype text/html.

    Adapted from setDefaults in Archetypes BasicSchema.
    """
    default_output_type = 'text/x-html-safe'
    mimetype = 'text/html'
    schema = instance.Schema()
    for field in schema.values():
        # We only need to do this for fields with one specific mimetype.
        if not shasattr(field, 'default_output_type'):
            continue
        if field.default_output_type != default_output_type:
            continue
        # only touch writable fields
        mutator = field.getMutator(instance)
        if mutator is None:
            continue
        base_unit = field.getBaseUnit(instance)
        if base_unit.mimetype == mimetype:
            continue
        # If content has already been set, we respect it.
        if base_unit:
            continue
        default = field.getDefault(instance)
        args = (default,)
        kw = {'field': field.__name__,
              '_initializing_': True}
        kw['mimetype'] = mimetype
        mapply(mutator, *args, **kw)



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _types
"""

# Python compatibility:
from __future__ import absolute_import

# Zope:
from Products.CMFCore.utils import getToolByName

# Logging / Debugging:
import logging

__all__ = [
        'setVersionedTypes',
        ]


def setVersionedTypes(portal, additional_types, logger=None):
    """
    Füge die übergebenen Typen (portal_type) der Änderungsverfolgung hinzu

    Achtung: um eine nützliche diff-Ausgabe zu erhalten, muß für jeden
    betroffenen Typ ein Diff-Tool konfiguriert sein! (siehe ../diff_tool.xml)
    """
    # siehe:
    # - http://www.uwosh.edu/ploneprojects/docs/how-tos/how-to-enable-versioning-history-tab-for-a-custom-content-type/
    # - https://docs.plone.org/4/en/manage/upgrading/version_specific_migration/p40_to_p41_upgrade.html#id7

    if logger is None:
        logger = logging.getLogger('setVersionedTypes')

    portal_repository = getToolByName(portal, 'portal_repository')
    old_policies = portal_repository.getPolicyMap()
    versionable_types = list(portal_repository.getVersionableContentTypes())
    for type_id in additional_types:
        if type_id not in versionable_types:
            logger.info('adding %(type_id)r to versionable types ...', locals())
            versionable_types.append(type_id)
        else:
            policies = old_policies.get(type_id, None)
            logger.info('versioning policies for %(type_id)r: %(policies)r', locals())

    logger.info('saving additions ...')
    portal_repository.setVersionableContentTypes(versionable_types)



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _uid
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types
from six import text_type as six_text_type

# Zope:
from Products.CMFCore.utils import getToolByName

# Local imports:
from visaplan.plone.tools._have import HAS_KITCHEN

if HAS_KITCHEN:
    # visaplan:
    from visaplan.kitchen.spoons import generate_uids
else:
    generate_uids = None

# Logging / Debugging:
from pdb import set_trace
from visaplan.tools.debug import pp

__all__ = [
        'make_distinct_finder',
        'make_uid_setter',
        'make_uid_collector',
        ]


def make_distinct_finder(**kwargs):
    """
    Erzeuge eine Funktion, die ein per UID angegebenes Objekt findet;
    bei uneindeutigem Ergebnis, oder wenn nicht gefunden, tritt ein Fehler auf.
    """
    if 'catalog' not in kwargs:
        context = kwargs.pop('context')
        catalog = getToolByName(context, 'portal_catalog')
    else:
        catalog = kwargs.pop('catalog')
    logger = kwargs.pop('logger')

    def find_one(**kwargs):
        """
        Finde exakt ein Objekt und gib das "brain" zurück.
        Derzeit wird nur nach UIDs gesucht ...
        """
        if 'uid' in kwargs:
            uid = kwargs.pop('uid')
        elif 'uid' in kwargs:
            uid = kwargs.pop('UID')
        else:
            raise ValueError('UID/uid needed; got %(kwargs)s', locals())
        query = {'UID': uid}
        brains = catalog(query)
        if not brains:
            raise ValueError('UID %(UID)r not found!', locals())
        elif brains[1:]:
            number = len(brains)
            raise ValueError('%(number)d hits for UID %(UID)r;'
                             ' *one* expected!',
                             locals())
        return brains[0]

    return find_one


def make_uid_setter(**kwargs):
    """
    Erzeuge eine Funktion, die die UID einer bekannten Ressource setzt;
    es wird nichts neu erzeugt
    """
    if 'catalog' not in kwargs:
        context = kwargs.pop('context')
        catalog = getToolByName(context, 'portal_catalog')
    else:
        catalog = kwargs.pop('catalog')
    logger = kwargs.pop('logger')

    find_one = make_distinct_finder(catalog=catalog, logger=logger)

    # Vorgabewerte:
    optional = kwargs.pop('optional', False)
    shortcircuit = kwargs.pop('shortcircuit', False)
    if kwargs:
        logger.error('make_uid_setter: unused arguments! (%(kwargs)r)', locals())

    if 'portal' not in kwargs:
        if 'context' in kwargs:  # might have been popped above already
            context = kwargs.pop('context')
        site = getToolByName(context, 'portal_url')
    else:
        site = kwargs.pop('portal')

    def visible_path(s):
        """
        >>> visible_path('/gkz/meine-kurse')
        '/meine-kurse'
        """
        assert s.startswith('/')
        liz = s.split('/')
        del liz[1]
        return '/'.join(liz)

    def set_uid(uid_new, path, uid_old=None,
                optional=optional,
                shortcircuit=shortcircuit):
        """
        uid_new -- die neue UID (jedenfalls benötigt)
        path -- ein Pfad, oder eine Sequenz von Pfaden (benötigt)
        uid_old -- optional; darf, wenn angegeben, nicht gleich uid_new sein

        optional -- das übergebene Dings ist optional; wenn es fehlt, ist das
                    nicht schlimm (aber wenn es da ist, soll es die angegebene
                    UID bekommen)
        shortcircuit -- Wenn mehrere Pfade angegeben wurden, nach dem ersten
                        Treffer aufhören zu suchen; ansonsten wird auch
                        geprüft, ob die anderen ebenfalls da sind, und dies
                        ggf. protokolliert
        """
        if isinstance(path, six_string_types):
            paths = [path]
        elif path:
            paths = tuple(path)
        else:
            paths = []
        brain_new = find_one(uid=uid_new)
        if uid_old:
            if uid_old == uid_new:
                raise ValueError('Different UIDs expected, got: %(uid_new)r'
                                 % locals())
            brain_old = find_one(uid=uid_old)
            if brain_old and brain_new:
                logger.fatal('set_uid: Both old (%(uid_old)r) and new '
                             '(%(uid_new)r) UIDs founc!', locals())
                return False
            o_old = brain_old.getObject()
            logger.info('Setting UID of %(o_old)r to %(uid_new)r'
                        ' (was: %(uid_old)r', locals())
            o_old._setUID(uid_new)
            return True

        if brain_new:
            path_new = visible_path(brain.getPath)
            if paths:
                if path_new in paths:
                    logger.info('Found UID %(uid_new)r in expected path'
                                ' %(path_new)r', locals())
                else:
                    logger.warn('Found UID %(uid_new)r in UNEXPECTED path'
                                ' %(path_new)r (expected: %(paths)s)', locals())
            else:
                logger.info('Found UID %(uid_new)r in path'
                            ' %(path_new)r (no expectations specified)',
                            locals())
            return True

        # keine Treffer über UIDs; jetzt nach Pfaden suchen:
        done = False
        for pa in paths:
            logger.info('seeking %(pa)r ...', locals())
            o = site.restrictedTraverse(pa)
            if o:
                if done:
                    logger.warn('ignoring %(o)r', locals())
                else:
                    logger.info('setting UID of %(o)r to %(uid_new)r', locals())
                    o._setUID(uid_new)
                    done = True
                    if shortcircuit:
                        break

        if done:
            return True
        elif optional:
            logger.info('Nothing found (new uid: %(uid_new)r,'
                        ' paths: %(paths)r',
                        locals())
            return False
        else:
            logger.error('Nothing found! (new uid: %(uid_new)r,'
                         ' paths: %(paths)r',
                         locals())
            return False

    return set_uid


def make_uid_collector(getbyuid, transform, extract=generate_uids,
                       theset=None,
                       expect='brain',
                       **kwargs):
    """
    Erzeuge eine Funktion, die rekursiv UIDs extrahiert und in einem Set
    sammelt; Rückgabewert ist ein Tupel (Funktion, Set).

    getbyuid - eine Funktion, die für eine gegebene UID das jeweilige
               Katalogobjekt zurückgibt
    extract - eine Funktion, die aus dem (ggf. durch die transform-Funktion
              ergänzten) Text die UIDs erzeugt
    transform - eine Transformationsfunktion, die den "rohen" Text expandiert
                und Einbettungen von Ressourcen auswertet
    theset - das Set (ggf. neu erzeugt)
    expect - einziger implementierter Wert: 'brain'

    Nur als benannte Option:

    debug_uids - eine Menge/Sequenz von UIDs, bei denen set_trace aufgerufen
                 werden soll. In diesem Fall wird eine spezielle Version der
                 Funktion erzeugt, die spezielle Debugging-Informationen
                 vorhält.
    debug_label - zur Information, welcher UID-Collector die aktuelle
                  Unterbrechung ausgelöst hat

    ACHTUNG:
    - die Vorhaltung des kompletten Rohtexts im Indexobjekt entspricht nicht
      den "best practices" für Plone;
    - es wird hier hartcodiert bislang nur der Wert des Felds "text" untersucht;
    - bei Dexterity-Objekten würde sich hier die Notwendigkeit diverser
      Änderungen ergeben!
    """
    if theset is None:
        theset = set()

    if expect != 'brain':
        raise ValueError('expect=%(expect)r not implemented!' % locals())

    if extract is None:
        raise ValueError('No extract function given!')

    def collect_uids(brain):
        """
        Rekursives Unterprogramm zur Extraktion von UIDs

        Argumente:
        - brain - ein Katalogeintrag
        """
        if not brain:
            return
        myuid = brain.UID
        if myuid in theset:
            return
        theset.add(myuid)
        text = brain.getRawText
        if not text:
            return
        elif transform is None:
            pass
        else:
            if isinstance(text, six_text_type):
                space = u' '
            else:
                space = ' '
            text += space + transform(text)
        for uid in extract(text=text):
            if uid in theset:
                continue
            collect_uids(getbyuid(uid))

    debug_uids = kwargs.pop('debug_uids', None)
    if not debug_uids:
        return collect_uids, theset
    debug_label = kwargs.pop('debug_label', None) or None
    if debug_label is None:
        headline = 'WATCHED CASE: uid %(uid)r found!'
    else:
        headline = ('WATCHED CASE (%(debug_label)s): uid %%(uid)r found!'
                    ) % locals()

    def collect_uids_2(brain, depth=0, stack=None):
        """
        Rekursives Unterprogramm zur Extraktion von UIDs

        Argumente:
        - brain - ein Katalogeintrag
        """
        if not brain:
            return
        myuid = brain.UID
        if myuid in theset:
            return
        theset.add(myuid)
        text = brain.getRawText
        if not text:
            return
        elif transform is None:
            pass
        else:
            if isinstance(text, six_text_type):
                space = u' '
            else:
                space = ' '
            text += space + transform(text)
        if stack is None:
            stack = [myuid]
        for uid in extract(text=text):
            if uid in theset:
                continue
            if uid in debug_uids:
                ppargs = [
                    headline % locals(),
                    ('myuid:', myuid),
                    ('depth:', depth),
                    ('stack:', stack),
                    ]
                if isinstance(debug_uids, dict):
                    uidinfo = debug_uids[uid]
                    ppargs.insert(1, ('UID-Info:', uidinfo))
                if 0 and 'debug collect_uids_2':
                    pp(ppargs)
                    set_trace()
            collect_uids_2(getbyuid(uid), depth+1, stack+[uid])

    return collect_uids_2, theset




# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _query
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types

# Zope:
from Products.CMFCore.utils import getToolByName

# Local imports:
from visaplan.plone.tools._idxnames import getExcludeFromSearch

# Logging / Debugging:
import logging

__all__ = [
        'make_query_extractor',
        'iterate_query',
        'getAllLanguages',
        ]


def make_query_extractor(context, do_pop=True):
    def extract_query(dic):
        """
        Extrahiere die bekannten Query-Argumente für die Katalogsuche
        und gib den Extrakt zurück
        """
        query = {}
        if do_pop:
            pop = dic.pop
        else:
            pop = dic.get

        if getExcludeFromSearch not in dic:
            query[getExcludeFromSearch] = False  # WIP: change index name
        else:
            val = pop(getExcludeFromSearch)
            if val is not None:
                # or bool(val)? what values do we expect (or support)?
                query[getExcludeFromSearch] = int(val)  # WIP: change index name
        # optionale Argumente ohne Vorgabewert:
        for name in [
            'portal_type',
            'getCustomSearch',
            'path',
            ]:
            if name in dic:
                val = pop(name)
                if isinstance(val, six_string_types):
                    val = [val]
                query[name] = val
        # Vorgabe: alle aktiven Sprachen
        Language = pop('Language', None)
        if Language is None:
            Language = getAllLanguages(context)
        # https://docs.plone.org/develop/plone/searching_and_indexing/query.html#bypassing-language-check:
        if isinstance(Language, six_string_types) and Language != 'all':
            Language = [Language]
        query['Language'] = Language
        return query
    return extract_query


def iterate_query(func, **kwargs):
    """
    func -- eine Funktion, die ein Katalogobjekt entgegennimmt und einen
            Wahrheitswert zurückgibt
    limit -- um die Bearbeitung auf <N> Objekte zu beschränken (zur
             Entwicklung)
    period -- wenn eine Zahl übergeben, werden die Änderungen nach je <period>
              Änderungen gesichert
    logger, catalog, context -- wie üblich; context wird jedenfalls benötigt

    sonstige benannte Argumente werden an --> make_query_extractor(context)
    übergeben
    """
    logger = kwargs.pop('logger', None)
    if logger is None:
        logger = logging.getLogger('reindex')
    catalog = kwargs.pop('catalog', None)
    if catalog is None:
        context = kwargs.pop('context')
        catalog = getToolByName(context, 'portal_catalog')
    limit = kwargs.pop('limit', None)
    period = kwargs.pop('period', None) or None

    extract_query = make_query_extractor(context)
    query = extract_query(kwargs)
    logger.info('iterate_query: query args are:\n  %s',
                '\n  '.join(['%r=%r' % tup
                             for tup in query.items()
                             ]))
    i = 0
    if period is not None:
        transaction.begin()
    try:
        for brain in catalog(query):
            if func(brain):
                i += 1
                if limit is not None and i >= limit:
                    break
                if period is None:
                    continue
                if not i % period:
                    logger.info('committing after %(i)r changes', locals())
                    transaction.commit()
    finally:
        if (period is not None and
            i % period
            ):
            logger.info('final commit after %(i)r changes', locals())
            transaction.commit()


def getAllLanguages(context, exclude=[]):
    """
    Zur Suche nach allen Sprachen (Index: "Language"),
    um keine Objekte zu verfehlen
    """
    language_tool = getToolByName(context, 'portal_languages')
    langs = sorted([x[0] for x in language_tool.listSupportedLanguages()])
    if '' not in langs:
        langs.append('')
    if exclude:
        exclude = set(exclude)
        unknown = exclude.difference(langs)
        if unknown:
            raise ValueError("Can't exclude unknown language codes %s!",
                             sorted(unknown))
        langs = set(langs)
        langs.difference_update(exclude)
        if not langs:
            raise ValueError('Excluding %s, nothing remains!',
                             sorted(exclude))
    return sorted(langs)
